/*
 *  (C) COPYRIGHT International Business Machines Corp. 1995, 1996
 *
 * Filename : 49
 * Author   : Myron Flickner
 * Date     : 5/13/96 14:36:16
 * Version  : 1.9
 * Purpose  : Main program to build or populate a QBIC catalog database
 *
 */

static char sccsid[] = "QbMkDbs.cpp, qbicapi, dmb_12 5/13/96" ; 

#include <iostream.h>

#include "Qbic.h"
#include "QbAbsCls.hpp"
#include "QbImgDat.hpp"
#include "QbCatCls.hpp"
#include "QbCreate.hpp"

static int Usage(char *pgm)
{
  const char usage[] =
    "[-r] [-u] [-E] [-c catalog] [-d database [-f feature1 [-f feature2 ...]][-s sampler:feature] image1 [image2 ...]\n"
    "\t-r removes the record or sample from the catalog or sampler\n"
    "\t-u updates versus inserts the record\n"
    "\t-c catalog specifies the catalog name (defaults to \"test\")\n" 
    "\t-d database specifies the database name (defaults to \"QbicData\"\n"
    "\t-E erases (drops) the catalog or sampler ( use with caution) !\n" 
    "\t-f feature adds (removes if -r specified) the feature to the catalog\n"
    "\t-s sampler:feature adds (removes if -r specified) the sampler to the catalog\n" ;
  printf("%s %s", pgm, usage ) ; 
  return 0 ; 
}


#define NUMBER_FEATURES 128

int main( int argc, char **argv, char **envp )
{
  int i, numberOfFeatures, argcImage, rc ;
  char *catalogName = "test" ;
  const char *databaseName = NULL ;
  static QbFeatureClass *features[NUMBER_FEATURES] ;
  QbCatalogClass *catalog ;
  QbDatumClass featureList, samplerList ;
  char *sampler = NULL, *samplerFeature=NULL ; 
  QbGenericImageDataClass image ;
  Boolean remove = False, update=False, erase=False ; 
  const QbConnectClass *connection ;
  unsigned long listSize ;
  char *listPtr ;
  
  if( 1 == argc ) {
    Usage(argv[0]) ;
    exit(1) ;
  }
  
  numberOfFeatures = 0 ; 
  for( i=1; i < argc; i++ ) {
    if( '-' == argv[i][0] ) {
      switch ( argv[i][1] ) {
      case 'c':
        catalogName = argv[++i] ; 
        break ;
      case 'd':
        databaseName = argv[++i] ;
        break ;
      case 'E':
        erase = True ;
        break ; 
      case 'f':
        features[numberOfFeatures] = QbCreateFeatureClass( argv[++i] ) ;
        if(  features[numberOfFeatures] ) numberOfFeatures++ ;
        else
          fprintf(stderr, "Unknown feature %s\n", argv[i] ) ;
        break ;
      case 'r':
        remove = True;
        break ; 
      case 's':
        sampler = argv[++i] ;
        break ;
      case 'u':
        update = True ;
        break ; 
      default:
        Usage(argv[0]) ;
        break ; 
      }
    }
    else break ; // done processing flags 
  }

  argcImage = i ; 
  
  connection = QbGetConnection( "w", databaseName ) ;
  if( NULL == connection ) {
    fprintf(stderr, "Unable to get connection to database %s\n",
            databaseName ? databaseName : 
            (getenv("QbicDatabaseName") ?
             getenv("QbicDatabaseName") : "QbicData" ) ) ; 
    return 1 ;
  }
  
  databaseName = connection->GetDSName() ; // set the database we really use

  catalog = connection->CreateCatalogClass() ; 
  
  catalog->OpenCatalog(catalogName) ; 
  //
  // process any sampler definitions
  //
  if( sampler ) {  // -s samplename:featureClassName
    samplerFeature = strchr( sampler, ':' ) ;
    if( samplerFeature ) {
      *samplerFeature = '\0' ;
      samplerFeature++ ;
    } // otherwise samplerFeature is NULL

    rc = catalog->ListSamplers( samplerList ) ;
    samplerList.Get( listSize, listPtr ) ;
    if( strstr( listPtr, sampler ) ) {
      printf("Sampler %s already defined\n", sampler ) ;
    } else {
      if( remove )
        rc = catalog->DropSampler( sampler ) ;
      else 
        rc = catalog->CreateSampler( sampler, samplerFeature ) ;
      if( rc ) {
        fprintf(stderr, "Failure to %s sampler %s %s rc = %d\n",
                remove ? "remove" : "create", sampler,
                samplerFeature, rc ) ;
        exit(1) ; 
      }
    }
  } else {
    rc = catalog->CreateCatalog( catalogName ) ; 
    i = 0 ; // add feature from command line 
    while( features[i] ) {
      if( remove )
        rc = catalog->DeleteFeatureClass( features[i]->Type() ) ;
      else 
        rc = catalog->AddFeatureClass( features[i]->Type() ) ;
      delete features[i] ;
      features[i++] = NULL ; 
    }
    // get feature from catalog 
    // rc = catalog->ListFeatureClasses( featureList ) ;
    // featureList.Get( listSize, listPtr ) ;
    // printf("%s\n", listPtr ) ; 
  }
    
  for(i=argcImage; i < argc; i++ ) {
    if ( ! remove ) {
      rc = image.GetFromFile( argv[i] ) ;
      if ( rc ) {
        fprintf(stderr,"Unable to read file %s rc = %d\n", argv[i],rc ) ; 
        continue ;
      }
    }
    if ( sampler ) {
      if( remove ) 
        rc = catalog->DeleteSample( argv[i], sampler) ;
      else
        rc = catalog->AddSample( argv[i], sampler, image, NULL, update ) ;
    } else {
      if( remove )
        rc = catalog->DeleteRecord( argv[i] ) ;
      else 
        rc = catalog->AddRecord( argv[i], image, NULL, update ) ;
    }
    
    if( rc )
      fprintf(stderr, "%s failed rc = %d file %s \n",
              remove ? "Delete" : "Insert", rc, argv[i]) ;
    else
      printf("%s of file %s into database=%s catalog=%s successful\n",
             remove ? "Delete" : "Insert",
             argv[i], databaseName, catalogName ) ; 
  }
  catalog->CloseCatalog() ;

  if( erase ) {
    if( sampler )  catalog->DropSampler( sampler ) ; // erases all sampler info
    else  catalog->DropCatalog( catalogName ); // erases all feature info !
  }
  
  delete catalog ; 
  QbDestroyConnection( connection ) ;
  return rc ;
}
