/*
 *  (C) COPYRIGHT International Business Machines Corp. 1995, 1996
 *
 * Filename : 51
 * Author   : Myron Flickner 
 * Date     : 7/16/96 23:07:46
 * Version  : 1.7.1.2
 * Purpose  : A test program to do query-by-example using the QBIC API.  
 *            QbQBE runs two successive queries.  The first is a keyword
 *            query.  Its results are used as a restriction list for 
 *            input to the second query.
 */

static char sccsid[] = "QbQBE.cpp, qbicapi, dmb_12 7/16/96" ; 

#include "QbAbsCls.hpp"
#include "QbImgDat.hpp"
#include "QbCreate.hpp"
#include "QbQueryC.hpp"
#include "QbQueryS.hpp"
#include <limits.h>

static void Usage(char *pgm)
{
  const char usage[] =
    "  -i query-image [-c catalog] [-d database]"
    "  -f feature1 [-f feature2 ...] \n" 
    "\t[ -T restriction-keyword ] -[ n nresults ] [-s 0|1 ] \n\n"
    "\t-i query-image specifies the query image \n" 
    "\t-c catalog specifies the catalog name (defaults to \"test\")\n" 
    "\t-d database specifies the database name (defaults to \"QbicData)\"\n" 
    "\t-f feature to use in the main (second) query\n"
    "\t-T the single keyword to use as a prefilter \n"
    "\t-n number of results to return (from second query) \n"
    "\t-s 0 (run query on server) 1 (run query locally - default)\n";
  printf("%s %s", pgm, usage ); 
  exit(200); 
}

#define NUMBER_OF_FEATURES 12

int main( int argc, char **argv, char **envp )
{
  int                      i, nhits = 20, rc = 0, server = 1, imageParmSet = 0;
  int                      numberOfFeatures, numberResults;
  char                     *catalogName="test", *databaseName="QbicData";
  char                     *imageName=NULL, *queryString, *buffer;
  QbFeatureClass           *thisFeature;
  const char               *featureNames[NUMBER_OF_FEATURES], 
                           *keywordFeatureName = "QbTextFeatureClass";
  QbGenericImageDataClass  *queryByExampleImage[NUMBER_OF_FEATURES] ; 
  QbGenericDataClass       *keywordData=NULL;  
  QbDatumClass             datum, key, value;
  QbQueryClass             *queryInstance = NULL ;
  unsigned long            size, keySize;
  const QbKeyDatabaseClass *firstResults = NULL, *secondResults = NULL ;
  const QbConnectClass     *connection;
  QbDbIteratorClass        *iter = NULL ; 
  float                    distance;

  numberOfFeatures = 0 ; 

  // read in input parameters from command line
  for( i=1; i < argc; i++ ) {
    if( '-' == argv[i][0] && argv[i+1] ) {
      switch ( argv[i][1] ) {
      case 'c':
        catalogName = argv[++i] ; 
        break ;
      case 'd':
        databaseName = argv[++i] ;
        break ;
      case 'T':  // the keyword to use to build the restriction list
        keywordData = new QbGenericImageDataClass;
        // set the type, to say this is an explicitly given text string
        keywordData->SetType(QbGenericDataClass::picker);
        // set the keyword in the text description field of the query
        // image structure
        keywordData->SetTextDescription(argv[++i]);
        break;
      case 'f':
        thisFeature = QbCreateFeatureClass( argv[++i] ) ;
        if( thisFeature ) {
          featureNames[numberOfFeatures] = thisFeature->Type();
          numberOfFeatures++ ;
        }
        else
          fprintf(stderr, "Unknown feature %s\n", argv[i] );
        delete thisFeature;
        break;
      case 'i':
        imageName = argv[++i];
	imageParmSet = 1;
        break;
      case 'n':
        nhits = atoi(argv[++i]);
        break;
      case 's':
        server = atoi(argv[++i]);
        break;
      default:
        Usage(argv[0]);
        break; 
      }
    }
    else break; 
  }

  // if no image, features, or text keyword specified, tell the user and exit
  if (imageParmSet == 0 || numberOfFeatures == 0 )
    Usage(argv[0]);

  for( i=0; i < numberOfFeatures ; i++ ) {
    queryByExampleImage[i] = new QbGenericImageDataClass;
    // this will be a query by example, specifying a named example image. 
    // other choices are query by an example image given explicitly as pixel
    // image, query by image from a sampler, etc. 
    queryByExampleImage[i]->SetType(QbGenericDataClass::example);
    queryByExampleImage[i]->SetCatalog(catalogName);
    queryByExampleImage[i]->SetKey(imageName);
  }

  // connect to the database.  
  connection = QbGetConnection("r", databaseName) ;

  if( NULL == connection ) {
    fprintf(stderr,"Unabled to get connection to database %s\n",databaseName );
    exit( 1 ) ;
  }
  
  if( server ) 
    queryInstance = QbCreateQueryClass("QbQueryServerClass") ;
  else 
    queryInstance = QbCreateQueryClass("QbQueryClientClass") ;
  
  if( NULL == queryInstance ) {
    fprintf(stderr,"Unabled to create query class\n" ) ;
    rc = 2 ;
    goto cleanup ; 
  }
  
  if( keywordData ) { // run the first query if keywords given
    
    // build the keyword query 
    QbBuildQueryString(LONG_MAX,   // a big number so we get everything back
                       catalogName,// catalog name to search in
                       NULL,       // initial restriction list (there is none)
                       NULL,       // restriction list for parts 
                       1,          // number of features -- one in this case
                       &keywordFeatureName, // feature class names 
                       NULL,                // use defaults class parms
                       &keywordData,       // structure with query features
                       datum ) ;            // query string (in datum format) 
    
    // set char* query string from datum query string    
    datum.Get(size, queryString);  
    

    // run the first query
    firstResults = queryInstance->Evaluate( queryString ) ; 

    numberResults = 0;
    if(firstResults)  numberResults = firstResults->GetNumberOfKeys() ;
    printf("%d results returned in first query\n", numberResults ) ;
  }
  else
    printf("No text specified. Skipping first query\n") ; 

  // now build the query string for the second query.  It will use 
  // the results of the first query as a restriction list
  QbBuildQueryString(nhits,   // number of results to return
    catalogName,              // catalog name to search in 
    firstResults,             // restriction list for scenes
    NULL,                     // restriction list for parts (none)
    numberOfFeatures,         // number of features in the query
    featureNames,             // feature class names 
    NULL,                     // use default class parms
                              // structure with query features
    (QbGenericDataClass **) queryByExampleImage,  
    datum ) ;                 // query string (in datum format) 

  // set char* query string from datum query string    
  datum.Get(size, queryString);  
 
  // run the second query
  secondResults = queryInstance->Evaluate( queryString ) ;
  rc = queryInstance->GetReturnCode() ; 

  if( rc ) {
    printf("Query failed return code %d\n", rc ) ;
    goto cleanup ;
  }
    
  if( secondResults ) {
    numberResults = secondResults->GetNumberOfKeys() ;
    printf("%d results returned in second query\n", numberResults ) ;
  
    iter = secondResults->CreateIterator();
    
    // print out the final query results
    while( iter->More() ) {
      rc = iter->Next( key, value ) ;
      key.Get( keySize,buffer ) ;
      distance = secondResults->GetDistance(value) ;
      printf("%s %f\n", buffer, distance ) ; 
    }
  } else
    printf("0 results returned in second query\n" ) ;
  
  // clean up
cleanup:
  delete iter;
  delete keywordData;
  delete queryByExampleImage[numberOfFeatures-1];
  delete queryInstance;
  QbDestroyConnection( connection ) ; 

  return rc; 
}
