/*
 *  (C) COPYRIGHT International Business Machines Corp. 1995, 1996
 *
 * Filename : 65
 * Author   : M. Flickner and others
 * Date     : 5/21/96 16:46:28
 * Version  : 1.16
 * Purpose  : Define the abstract classes used for qbic extensibility
 *
 */

#ifndef __QbAbsCls_hpp
#define __QbAbsCls_hpp

#include <string.h>
#include "Qbic.h"
#include "QbUtil.h"

////////////////////////////////////////////////////////////////
//
// QBIC API objects generally carry a dynamic type.  This type
// information is accessible through the interface to QbBase.
//
// We need this dynamic typing information because this is a
// dynamically configurable system; new kinds of QbFeature types
// require corresponding new kinds of types.
//
////////////////////////////////////////////////////////////////

class QbBaseClass {
public:
  QbBaseClass() {} ;
  virtual ~QbBaseClass() {} ;

  //
  // The interface of particular queries to the QBIC drivers is
  // described in terms of a collection of C++ classes with interfaces
  // specified in this file.  QbInterfaceType enumerates these basic
  // kinds of interfaces.
  // 
  enum QbInterfaceType {
    Undefined = 0,
    DatumInterfaceType = 98732,
    GenericDataInterfaceType = 171001,
    FeatureInterfaceType = 331007,
    ConnectInterfaceType = 301397,
    DatabaseInterfaceType = 452398,
    KeyDatabaseInterfaceType = 501983,
    DatabaseIteratorInterfaceType = 543976,
    ParameterInterfaceType = 558797,
    CatalogInterfaceType = 754864,
    QueryInterfaceType = 877543
  };

  //
  // IType() returns the constant that defines the interface type as
  // defined in the QbInterfaceType enum.  
  //
  virtual const QbInterfaceType IType() const = 0;
  // 
  // Type() is required to return a character string identical to the
  // name of the derived class to enable the COERCE macro to safely cast
  // a pointer to an abstract class to a pointer to a derived class.
  //  
  virtual const char *Type() const = 0;

  //
  // check whether the type name returned by Type() is
  // the same as some given string
  //
  virtual Checktype(const char *s) const {
    return !strcmp(s,Type());
  }
};

//
//The COERCE macro compares the string returned by Type() with the
//passed class name. It returns NULL if the conversion is not allowed.
//Analogous to Eiffel's (safe) reverse assignment;
//
#define COERCE(Object,Type) (((Object)->Checktype(#Type))?((Type *)(Object)):0)

//
// Unexposed classes ( no .hpp file supplied ) used in QbCreate.cpp
// need to use this macro to define a constructor. 
//
#define NEW_FUNCTION( base, derived ) \
base *New##derived() { return new derived ; } 

////////////////////////////////////////////////////////////////
//
// We need a class to store key value pairs.  This is a full implementation
// that allows the caller to tell the class where the storage is comming
// from.  Storage will be deleted on a subsequent call to Set or 
// when the class is destroyed. 
// 
////////////////////////////////////////////////////////////////
class QbDatumClass : public QbBaseClass {

private:  
  QbDatumClass( const QbDatumClass &in ) ; // avoid any copies by users
  
public:
  enum storageEnum {leaveAlone, mallocStorage, newStorage, newArrayStorage}; 
  
protected:
  unsigned long size ;
  unsigned char *bytes ;
  char *format ;
  storageEnum bytesStorageFrom ;
  storageEnum formatStorageFrom ; 
  
  void CleanupBytes() {
    if( leaveAlone == bytesStorageFrom ) return ;
    if( mallocStorage ==  bytesStorageFrom )  free( bytes ) ;
    if( newStorage == bytesStorageFrom  ) delete bytes ;
    if( newArrayStorage == bytesStorageFrom ) delete [] bytes ; 
  }
  
  void CleanupFormat() {
    if(leaveAlone == formatStorageFrom) return ;
    if( mallocStorage ==  formatStorageFrom )  free( format) ;
    if( newStorage == formatStorageFrom  ) delete format ;
    if( newArrayStorage == formatStorageFrom ) delete [] format ; 
  }
  
public:
  
  QbDatumClass(): size(0), bytes(NULL), format(NULL),
  bytesStorageFrom(leaveAlone),formatStorageFrom(leaveAlone) {} 

  virtual ~QbDatumClass() {
    CleanupBytes() ;
    CleanupFormat() ; 
  }
  //
  // From QbBaseClass
  //
  virtual const QbInterfaceType IType() const { return DatumInterfaceType ; }
  virtual const char *Type() const { return "QbDatumClass" ; }
  //
  // Set the size of the buffer, a pointer to the buffer and tell the
  // class what to do with the storage on future calls to set or in the
  // destructor. 
  //
  virtual void Set(const unsigned long nArg, // size of the buffer in bytes
                   unsigned char *pArg,      // ponter to the buffer 
                   const storageEnum storageType = leaveAlone ) {
    CleanupBytes() ; // avoid leaks
    size = nArg ;
    bytes = pArg ;
    bytesStorageFrom = storageType ;
  }

//
// Get the size of the buffer and a pointer to the buffer 
//  
  virtual void Get(unsigned long &n, unsigned char *&p) const {
    n = size ; 
    p = bytes ;   
  }

  virtual void Get(unsigned long &n, char *&p) const {
    n = size ;
    p = (char *) bytes ;
  }

  virtual void Get(unsigned long &n, void *&p) const {
    n = size ;
    p = bytes ;
  }
  

  //
  //Format information can be used to export information about the
  //byte stream.  For example Qbic applications can use the format
  //information to export images features to a relational database.
  //The format information can be used to store the database column
  //name and feature type information.
  //
  virtual void SetFormatInfo(char *f, // format stored as a char string
                             const storageEnum storageType = leaveAlone)
  {
    CleanupFormat();
    format = f;
    formatStorageFrom = storageType ;
  } 

  //
  // Function to get the format string.
  //  
  virtual const char *GetFormatInfo() const { return format ; }

  //
  // There are times when it is useful to copy datums.  This assignment
  // operator allows for copies of the form dataout = datumin.  A copy
  // of both the data and the format information is made. 
  // 
  QbDatumClass& operator =(const QbDatumClass &in) {
    unsigned char *inBuf ;
    unsigned char *outBuf ;
    unsigned long size ;
    const char *fBuf ;
    char *foutBuf ; 
    
    Set(0,NULL) ;
    SetFormatInfo(NULL) ;

    in.Get( size, inBuf) ;
    if( size ) {
      outBuf = new unsigned char[size] ;
      memcpy( outBuf, inBuf, size) ;
      Set( size, outBuf, newArrayStorage) ;
    }
    if( ( fBuf = in.GetFormatInfo() ) != 0) {
      size = strlen(fBuf) + 1 ;
      foutBuf = new char[size] ;
      memcpy( foutBuf, fBuf, size ) ;
      SetFormatInfo( foutBuf, newArrayStorage) ;
    }
    return *this ; 
  }
};

class QbDatabaseClass ;
class QbDbIteratorClass ;
class QbKeyDatabaseClass ;
class QbCatalogClass ; 

//
// The purpose of this class is to setup the connection to
// the database or mount the file system, create a directory...
// dsName can be a database name or a file server pathname...
// This is also how one creates an instance of the database class
//  
class QbConnectClass : public QbBaseClass {
protected:
  char *connectMode, *dsName ;
  Boolean isConnected ;
public:
  QbConnectClass() : connectMode(NULL), dsName(NULL), isConnected(False) {}
  virtual ~QbConnectClass() {}

  // Connect sets dsName, connecteMode and isConnected 
  virtual int Connect( const char *dsNameArg, const char *modeArg ) = 0 ;
  // reverses the connect and frees any storage for dsName and connectMode 
  virtual int Disconnect() = 0 ; 

  // this is the architected way to create a database class
  virtual QbDatabaseClass *CreateDatabaseClass() const = 0 ;

  // this is the architected way to create a key database class
  virtual QbKeyDatabaseClass *CreateKeyDatabaseClass() const = 0 ;

  // this is the architected way to create a catalog class
  virtual QbCatalogClass *CreateCatalogClass() const = 0 ;

  virtual int IsConnected() const { return isConnected ; } 
  virtual const char *GetConnectMode() const { return connectMode ; }
  virtual const char *GetDSName() const { return dsName ; } 
} ;

//
// Abstract class to define a persistant storage class  
//
class QbDatabaseClass:  public QbBaseClass {
private:
  QbDatabaseClass( const QbDatabaseClass& ) ; // no copies
  QbDatabaseClass& operator =(const QbDatabaseClass&) ;
protected:
  const QbConnectClass &connection ;
public:
  QbDatabaseClass(const QbConnectClass &connectArg) :
    connection( connectArg ) { }
  virtual ~QbDatabaseClass() {} 

  const QbConnectClass &GetConnection() const { return connection ; }
  //
  // Opens a table or a file.
  //
  virtual int Open(const char *tableOrFileName) = 0 ;
  virtual int Close() = 0 ; // reverses the Open

  // Since Open creates a file/table if it doesn't exist
  // we need a way to delete a file/table.  Open the table/file then use
  // CloseAndRemove to delete it. 
  virtual int CloseAndRemove() = 0 ;
  
  // If there is transaction support this checkpoints all the updates, adds
  // and deletes.
  virtual int Commit() = 0 ; 
  
  // inserts a record only if the record doesn't exist
  virtual int Insert(const QbDatumClass &key, const QbDatumClass &value) = 0;

  // Updates a record if it exists.  If insert is False the update
  // fails if the record doesn't exist.  If insert is True update
  // inserts the record if it doesn't exist.
  virtual int Update(const QbDatumClass &key,
                     const QbDatumClass &value,
                     const int insert = False ) = 0;

  // reads a record
  virtual int Retrieve(const QbDatumClass &key, QbDatumClass &value) const = 0;

  // deletes a record
  virtual int Delete(const QbDatumClass &key) = 0;

  // This creates an iterator over the database. This returns NULL
  // if the database is not open.
  virtual QbDbIteratorClass *CreateIterator() const = 0 ;

  virtual const char *GetContainerName() const = 0 ;
} ;

//
//A class to return the results of a query or to define a list of
//keys.  The result is a collection of keys order first by distance
//then by key.  SetDistance should be called before Insert is called.
//
class QbKeyDatabaseClass: public QbDatabaseClass {
protected:
  Boolean isRanked ;
  unsigned long numberOfKeys ; // the number of elements in the database
                               // Inserts increments, Delete decrements
public:
  QbKeyDatabaseClass(const QbConnectClass &connect ) :
    isRanked(False), numberOfKeys(0), QbDatabaseClass(connect) {} 
  
  virtual ~QbKeyDatabaseClass() {}

  
  Boolean IsRanked() const { return isRanked ; }
  void SetRanked( const Boolean rank ) { isRanked = rank ; }

  unsigned long GetNumberOfKeys() const { return numberOfKeys ; } 

  // converts a key database into an ascii string
  // This could be a list of keys, the name of a persistant object...
  virtual int ToQueryString(QbDatumClass &datum) const = 0  ;
  
  // converts the ascii string from ToQueryString back into a key database
  virtual int FromQueryString(const QbDatumClass &datum) = 0 ; 
  
  virtual int SetDistance(QbDatumClass &data,
                          const float distance = -1 ) const = 0 ;
  
  virtual float GetDistance(const QbDatumClass &data ) const = 0 ;

  virtual int RetrieveMax( QbDatumClass &key, QbDatumClass &value ) const = 0 ;
} ;


////////////////////////////////////////////////////////////////
//
// Abstract class to define an iterator over a database
// functions map to OQS (Object Query Service) OMG TC doc 95.1.1
//
////////////////////////////////////////////////////////////////
class QbDbIteratorClass:  public QbBaseClass {
private :
  QbDbIteratorClass( const QbDbIteratorClass& ); // no copies
  QbDbIteratorClass& operator =(const  QbDbIteratorClass&) ;
protected:  
  char *const connectFilename ; 
public:
  QbDbIteratorClass( const QbDatabaseClass *db ) :
  connectFilename( QbStrDup( db->GetContainerName() )) {} ; 
  virtual ~QbDbIteratorClass() { delete connectFilename ; }
  virtual int More() = 0;
  virtual int Reset() = 0;
  
  enum nextEnum { FirstKey, NextKey, ThisKey } ;
  virtual int Next(QbDatumClass &key,
                   QbDatumClass &value,
                   nextEnum nextFlag = NextKey ) = 0 ;
} ;

#include "QbGenDat.hpp"
#include "QbPrmCls.hpp"

////////////////////////////////////////////////////////////////
//
// Abstract class to define a set of features
//
////////////////////////////////////////////////////////////////
class QbFeatureClass : public QbBaseClass {
  
public:
  QbFeatureClass() {} ;
  virtual ~QbFeatureClass() {} ;
  //
  //Features need to be converted to a byte stream.  ToByteString
  //will return heap storage that will be deleted when the datum is
  //destroyed or when Set is run using this datum.
  //  
  virtual int ToByteString( QbDatumClass &data ) = 0;
  virtual int FromByteString( const QbDatumClass &data ) = 0;
  //
  //Given generic data and parameters ComputeFeature will compute the
  //feature/indexing information.  The feature information can be
  //converted to/from a byte string via To/FromByteString and thus
  //stored using Insert in an QbDatabaseClass.  ComputeFeatures
  //returns 0 for success non-zero for failure.  Compute features
  //converts GenericData into features.
  //
  virtual int ComputeFeature(const QbGenericDataClass *data,
                             const QbParameterClass *parms ) = 0 ;


  //
  // During database population, some features require global
  //information such as statistics ( count, sum, sum^2) to be
  //computed.  This can be accomplished by passing in the current
  //state of the global information as a datum along with what
  //operation ( delete, insert) will be performed with this feature to
  //the database.  Normally, this works after features have been
  //imported via FromByteString or ComputeFeatures.  QbCatalogClass
  //will store this information in the database and call this function
  //on all Inserts, Updates and Deletes. The function,
  //SizeofGlobalFeatureInfo, returns the size of the global feature
  //info. When a feature is added to a catalog, the global storage is
  //set to zero.  The first unsigned long in datum must be the number
  //of records in the database.  We keep the information in the
  // database as an ascii string to avoid byte order dependancies. 
  // 
  enum GlobalFeatureInfoEnum { Insert, Delete, indexInsert, indexDelete } ;
  virtual int UpdateGlobalFeatureInfo( QbDatumClass &datum,
                                       const GlobalFeatureInfoEnum type ) const
  {
    unsigned long count=0, indexcount=0, size ;
    char *tmp ; 
    datum.Get( size, tmp ) ;
    if( '\0' != tmp[0] ) sscanf(tmp, "%lu %lu", &count, &indexcount ) ;
    if( type == Delete ) count-- ;
    else if (type == Insert ) count++ ;
    else if (type == indexDelete ) indexcount-- ;
    else if (type == indexInsert ) indexcount++ ;
    sprintf(tmp,"%010lu %010lu",count, indexcount ) ; 
    return 0 ;
  }

  virtual unsigned long SizeofGlobalFeatureInfo() const
  { return 22 ; } // 10 digits + space + 10 digits + '\0' 
  
  //
  // Distance computes the distance between this feature instance (
  // typically the query feature ) and the another feature instance using
  // any query specific parameters parms. The abort distance argument can
  // be used to speed up distance compassions. For example, say we issue
  // a a query and are looking for 20 hits. As we scan the database we
  // keep an order lists of the best 20 hits. By setting abortDistance to
  // the distance of the 20'th hit the Distance is free to abort the
  // computation whenever the distance exceeds this limit since we know
  // that it will not be in the top 20. For complicated distance
  // functions this fastpath give considerable savings.  The returned distance
  // must be scaled between 0.0 and 1.0 to enable multiple feature searches.
  //
  virtual fastfloat Distance(const QbFeatureClass *feature,
                             const QbParameterClass *parms,
                             const float abortDistance) = 0;
  //
  // For cost based query optimization the optimizer needs to know long
  // does it take to compute the distance between to points or how long
  // it take to execute this filtering process. Then the query optimizes
  // can rewrite the query for fewer distance evaluations or even decide
  // when to do an exact parametric search before the ranking. Even
  // though it is platform specific execution cost returns time in
  // seconds. Since we currently don't have a cost based optimizer it
  // doesn't matter what value is returned.
  //  
  virtual float
  ExecutionCostForDistanceFunction(const QbFeatureClass *feature,
                                   const QbParameterClass *parms,
                                   const float abortDistance) = 0 ;
  
  virtual float
  ExecutionCostForFilterFunction(QbDatabaseClass &database,
                                 const QbKeyDatabaseClass *list,
                                 const unsigned long queryResultsRequested,
                                 const char *catalogName, 
                                 const QbParameterClass *parms ) = 0 ;

  //
  //Filtering the database is akin to an index search of a
  //database. The passed arguments are a connected database, an
  //optional key sublist of this database, and the query
  //parameters. FilterResults can return a ranked or an un-ranked
  //list. If the list is un-ranked then the caller calls Distance for
  //each item in the list to determine the finally ranking. If the
  //list is ranked no further processing is needed. FilterResult is a
  //required interface but it not required for the implementation.
  //Returning a return code of -1 tells the caller that this feature
  //has no filter function. In this case the caller runs the query
  //using table scan and the Distance functions.  Likewise, if filter
  //returns a ranked list the Distance function is not required.
  //
  virtual int FilterResult(QbDatabaseClass &database,
                           const QbKeyDatabaseClass *list,
                           const unsigned long queryResultsRequested,
                           const char *catalogName, 
                           const QbParameterClass *parms,
                           QbKeyDatabaseClass &results ) = 0;
  
  //
  // The name of the QbGenericDataClass that this feature supports.
  //
  virtual const char *GenericDataClassname() const = 0;

  //
  // The name of the QbParameterClass that this feature suppport.
  //
  virtual const char *ParameterClassname() const = 0;

  //
  // function to return the default parameters
  //
  virtual const QbParameterClass *GetDefaultParameters() const = 0 ;

  //
  //Some features may need to update weights and such things using the
  //global information related to a specific catalog.  This can be
  //used to scale each feature with its variance or even compute true
  //Mahalanobis distance.
  //  
  virtual int ParameterUpdateFromGlobalInfo( QbParameterClass &parms,
                                             const QbDatumClass &globalInfo 
                                             ) const {  return 0 ; }
};

//
//Abstract class to evaluate a query - there are two instances of the
//class QbQueryClientClass and QbQueryServerClass.
//
class QbQueryClass: public QbBaseClass {
protected:
  QbKeyDatabaseClass *results ;
  int returnCode ; 
public:
  
  QbQueryClass() : results(NULL), returnCode(0) {} ;
  virtual ~QbQueryClass() { delete results ; }
  virtual GetReturnCode() const { return returnCode ; }
  
  virtual const QbKeyDatabaseClass * Evaluate( const char *query ) = 0 ;

} ; 

#endif

