/*
 *  IBM Confidential
 *  (C) COPYRIGHT International Business Machines Corp. 1995
 *
 * Filename : 66
 * Author   : M. Flickner, T. Donn
 * Date     : 5/21/96 13:54:37
 * Version  : 1.6
 * Purpose  : Interface to create/populating qbic databases
 *
 */
#ifndef __QbCatCls_hpp
#define  __QbCatCls_hpp

#include "QbString.hpp"
#include "QbAbsCls.hpp"

class QbCatalogClass : public QbBaseClass {

  friend class QbQueryServerClass ; 
  
protected:
  QbStringClass catalogName;//The name of the catalog
  QbStringClass dbClassName;//The C++ class name of the database class to use  
  QbDatabaseClass *db ;
  const QbConnectClass &connection ; 

//
// Next three functions will become private in the future !!!
//  
public: // soon to be private: !!!!!
  //
  // Return the name of the table/file that contains the feature
  //
  int GetFeatureTableName(const char *feature, // name of feature 
                          QbStringClass &name  // returned table name
                          ) const ;

  //
  // Return the name of the table/file that contains the feature index data
  //
  int GetFeatureIndexTableName(const char *feature, // name of feature 
                               QbStringClass &name  // returned table name
                               ) const ;

  //
  // Return the name of the table/file that contains the sampler
  //
  int GetSamplerTableName(const char *sampler, // name of sampler
                          QbStringClass &name  // returned table name
                          ) const ;

public:

  QbCatalogClass(const QbConnectClass &) ;
  virtual ~QbCatalogClass() ;

  // from QbBaseClass
  virtual const QbInterfaceType IType() const { return CatalogInterfaceType ;} 
  virtual const char *Type() const { return "QbCatalogClass" ; }

  //
  //Create the catalog in the persistent storage. A container will be
  //created in the persistent storage for the global catalog info such
  //as feature names, global feature information...
  virtual int CreateCatalog ( const char *catalogNameArg ) ; 

  //
  // Drop (delete) the catalog from the persistent storage
  //
  virtual int DropCatalog ( const char *catalogNameArg ) ;

  //  
  // Open the catalog for updates, inserts, deletes
  //
  virtual int OpenCatalog ( const char *catalogNameArg ) ;

  //
  // close the catalog
  //
  virtual int CloseCatalog ( ) ;

  //
  // gets database class name
  //
  virtual const char *GetDatabaseClassName();

  //
  // The following functions are valid only after the Catalog is opened.
  // Add a new feature class to the catalog A container will be created
  // in the persistent storage for the feature data
  //
  virtual int AddFeatureClass (const char *featureClassName) const ;

  //
  // Delete a feature class from the catalog
  //
  virtual int DeleteFeatureClass (const char *featureClassName) const ;

  //
  // List current feature class names
  //
  virtual int ListFeatureClasses (QbDatumClass &featureList) const ;

  //
  // Add an object to the catalog Features will be computed and saved in
  // the persistent storage for all current feature classes in the
  // catalog
  //
  virtual int AddRecord (const char *keyArg, // input object key
                         const QbGenericDataClass &data,
                         // array of parameter classes - one for each
                         // feature - must be in order of features as
                         // returned by ListFeatureClasses.
                         // If NULL use feature class
                         // default parmaeter class.
                         const QbParameterClass *parms, 
                         // update instead of insert
                         const int update = False ) const ;
  //
  // Delete an record from the catalog.
  //
  virtual int DeleteRecord (const char *keyArg) const ;

  //
  // Set the feature with information from the persistant store
  //
  virtual int GetFeature (const char *keyArg,
                          QbFeatureClass &feature) const ; 

  //
  // Get the record ( raw data ) from persistant store
  //
  virtual int GetRecord( const char *keyArg,
                         const char *featureClassname,
                         QbDatumClass &data  ) const ; 
  //
  // Number of objects
  //
  virtual unsigned long NumberOfRecords(const char *featureClassname) const ;

  //
  // Number of index database records
  //
  virtual unsigned long NumberOfIndexRecords(const char *featureClassname) const ;

  //
  // Return global feature informataion in datum
  //
  virtual int GetGlobalFeatureInfo( const char *featureName,
                                    QbDatumClass &datum ) const ;

  //
  // Save the global feature information in datum
  //
  virtual int PutGlobalFeatureInfo( const char *featureName,
                                    QbDatumClass &datum ) const ;
  
  //
  // Function to define subparts of a query object.  In Qbic this allows
  // for object ( Region of Interest) definition to be defined.
  //  
  virtual int DefineSubpart( const char *parentKey, // key to base information
                             const char *subpartKey, // key for subpart
                             const QbGenericDataClass &subpart) const ;
  //
  // Return the subpart definition given the subpart key
  //
  virtual int GetSubpartDefinition( const char *subpartKey,
                                    QbGenericDataClass &subpart ) const ;
  //
  // Function to get the keys of all the subpart definitions
  //  
  virtual int GetSubpartKeys( const char *parentKey,
                              QbDatumClass &subpartKeys ) const ;
  //
  // Given a subpart key return the key to the parent
  //
  virtual int GetParentKey( const char *subpartKey,
                            QbDatumClass &parentKey ) const ; 

/////////////////////////////////////////////////////////////////////////  
//
// The following functions are for sampler support.  They do not
// require a catalog be opened before use.
//
/////////////////////////////////////////////////////////////////////////
  
  //
  // Create a new sampler table/dbm file
  //
  virtual int CreateSampler (const char *samplerName,
                             const char *featureClassName) const ;
  //
  // Drop a Sampler table/dbm file
  //
  virtual int DropSampler (const char *samplerName ) const ;

  //
  // List current sample names and their feature classes
  //
  virtual int ListSamplers (QbDatumClass &samplerList ) const ; 

  //
  // Add a sample to the catalog Feature will be computed and saved in
  // the persistent storage as a sampler in the catalog
  //
  virtual int AddSample (const  char *keyArg,
                         const char *samplerName,
                         const QbGenericDataClass &data,
                         const QbParameterClass *parms,
                         const int update=False 
                         ) const ;        // input object key
  //
  // Delete a sample from a sampler
  //
  virtual int DeleteSample (const char *key,
                            const char *samplerName ) const  ;

  // 
  // Get a sample feature from a sampler
  //
  virtual int GetSampleFeature (const char *key,
                                const char *samplerName,
                                QbFeatureClass &feature ) const ;
  
  //
  // Get sampler data from a catalog
  //
  virtual int GetSampleRecord( const char *key,
                               const char *samplerName,
                               QbDatumClass &datum ) const ; 
  //
  // Number of samples in a sampler
  //
  virtual unsigned long NumberOfSamples(const char *samplerName) const ;

} ;

#endif
