/*
 *  IBM Confidential
 *  (C) COPYRIGHT International Business Machines Corp. 1995
 *
 * Filename : 76
 * Author   : Denis Lee
 * Date     : 4/29/96 21:52:55
 * Version  : 1.4
 * Purpose  : Class definition for the storage and import/export of parameters.
 *
 */

#ifndef __QbPrmCls_hpp
#define __QbPrmCls_hpp

#include "assert.h"

////////////////////////////////////////////////////////////////
//
// Class to define parameters assocated with a feature/query
//
////////////////////////////////////////////////////////////////
class QbParameterClass : public QbBaseClass  {
private:
  QbParameterClass(const QbParameterClass&) ; // no copies


protected:
  float classWeight; // class weight factor
  unsigned int count;    // # of features in the class
  char  **featureCodes;  // A character description of each class featue
  fastfloat *weights;    // The floating point weight for each feature

public:
  QbParameterClass() : classWeight(1.0), count(0),
    featureCodes(NULL), weights(NULL) {} ;

  virtual ~QbParameterClass() {
    unsigned int i ;
    for( i=0; i < count ; i++ ) delete featureCodes[i]  ;
    delete featureCodes ;
    delete weights ;
  }

  //
  // Copy assignment operator i.e. a = b, - copy all of the data
  //
  QbParameterClass& operator =(const QbParameterClass &rhs ) {
    int rc ;
    unsigned int i ; 

    if( 0 == GetCount() ) 
      rc = AllocateStorage( rhs.GetCount() ) ;

    // give up if sizes don't match 
    assert(  GetCount() == rhs.GetCount() ) ; 
    
    SetClassWeight( rhs.GetClassWeight() ) ;

    for( i=0; i<count;i++ ) {
      SetFeatureCodes(i,rhs.GetFeatureCodes(i)) ;
      SetWeights(i, rhs.GetWeights(i) ) ; 
    }
    return *this ;
  }
  
  //
  // From QbBaseClass
  //
  virtual const QbInterfaceType IType() const
      { return ParameterInterfaceType ; }
  virtual const char *Type() const { return "QbParameterClass"; }

  int ToQueryString(QbDatumClass &datum) const;
  int FromQueryString(const QbDatumClass &datum);
  
  //
  // member functions to allocated the storage needed.  Ideally this would
  // be in the constructor.  However, to remain consistant with the
  // dynamic binding in QbBaseClass we have to have a no argument
  // constructor. 
  //
  int AllocateStorage(const unsigned int size) {
    unsigned int i ;  
    if( count ) return -1 ; // only allow this to be run once. 
    count = size ;
    featureCodes = new char*[size] ;
    weights = new fastfloat[size] ;
    
    for( i = 0; i < count ; i++ ) {
      featureCodes[i] = NULL;
      weights[i] = 0.0 ; 
    }
    classWeight = 1.0 ;
    return 0 ; 
  }
  
  unsigned int GetCount() const  { return count ; } 
  
  fastfloat GetClassWeight() const  { return classWeight; }
  void SetClassWeight( fastfloat x ) { classWeight = x ; }
  
  char **GetFeatureCodes() const { return featureCodes ; } 

  char *GetFeatureCodes(const unsigned int i) const {
    return i < count ? featureCodes[i] : (char *) NULL ; }

  void SetFeatureCodes( const unsigned int i, const char *value ) {
    if ( i < count ) {
      delete featureCodes[i] ;
      featureCodes[i] = QbStrDup( value ) ;
    }
  }
  
  fastfloat* GetWeights() const { return weights ; } 

  fastfloat GetWeights(const unsigned int i) const 
  { return i < count ? weights[i] : 0.0  ; }

  void SetWeights( const unsigned int i, const fastfloat value )
  { if ( i < count) weights[i] = value ; } 

} ;

#endif
