#!/bin/ksh
wish << \#EOF
set viewer "xv"

# This script was tested on AIX 4.1.4
# running Tcl version 7.5 alpha 2, and Tk version 4.1 alpha 2.
# Older versions of these packages may cause some functions
# to be unsupported, such as shell argument parameter passing
# and menu widget options. 
# (C)July 30,1996 RNYU@IBM-ARC
# --------INITIALIZATION----------
# The following variables can be set in the .qbictcl file:
# 1) QbicImageBase - the images directory
# 2) QbicDataPath - the database directory (database name)
# 3) QbicCatalogName - the default catalog name
# 4) QbicThumbPath - the target directory for thumbnails generation
# 5) imageExt - the list of images file extensions to process

set logfile "/tmp/QbCreateCat.log"

# Output usage message

if { ($argc >= 2) || ( ($argc == 1) &&  ![string compare $argv "-help"] )} {
    # puts stdout "argc=$argc"
    # puts stdout "argv=$argv"
    puts stdout { Usage: QbCreateCat.tcl [defaults file path] }
    puts stdout { You may specify the directory path to the defaults file .qbictcl file as argument.}
    exit
}


# Set the defaults file - this file should be named ".qbictcl"
set defltfile ".qbictcl"
set defpath [pwd]


# If the defaults file path is specified 
if {$argc == 1} {

    set defpath [lindex $argv 0]
    # puts stdout $defpath

    # Read the defaults file for variable assignments
    # The existence of .qbictcl in the user's home directory is 
    # checked.  If the file does not exist, only some defaults are assigned.

    if [file exists $defpath/$defltfile] {
      puts stdout "Reading $defpath/$defltfile"
      source $defpath/$defltfile
    } else {
        puts stdout "Error: $defltfile not found in $defpath"
        exit
    }
}


 # If no default file path is specified or the path specified 
 #  yielded no file,  select defaults will be assigned
if { ($argc == 0) || ![file exists $defpath/$defltfile] } { 

    # If argc==0, go ahead and read the default file if it exists
    if [file exists $defpath/$defltfile] {
	source $defpath/$defltfile
    }

    # Below, the locally assigned Qbic variables are checked for their existence.
    # "Local" means assigned within tclsh, ie., in the optional .qbictcl file
    #  which allows values to be assigned to override the Qbic environment variables.
    # If no values as assigned to these local variables, their corresponding
    # environment variables are used.  If no corresponding environment variables
    # are set, then default values are assigned within this tcl script.


    # Set the base path to the current directory 

    if ![info exists QbicImageBase] {
        if [catch {![info exists $env(QbicImageBase)]} rc] {
            if {[string compare [string range $rc 0 9] "can't read"] == 0} {
                set basePath [pwd]
            }
        } else {
            set basePath $env(QbicImageBase)
        }
    } else {
	set basePath $QbicImageBase
    }

    # Set the default image files extensions to include (all) formats
    if ![info exists imageExt] {
      set imageExt {tif tiff gif pcx jpg jpeg jfif bmp TIF TIFF GIF PCX JPG JPEG JFIF BMP}
    }
    
    # Set the default database name
    if ![info exists QbicDataPath] {
        if [catch {![info exists $env(QbicDataPath)]} rc] {
            if {[string compare [string range $rc 0 9] "can't read"] == 0} {
                set dbName ""
            }
        } else {
            set dbName $env(QbicDataPath)
        }
    } else {
	set dbName $QbicDataPath
    } 

    # Set the default catalog name

    if ![info exists QbicCatalogName] {
        if [catch {![info exists $env(QbicCatalogName)]} rc] {
           if {[string compare [string range $rc  0 9] "can't read"] == 0} {
               set catalogName "test"
           }
        } else {
            set catalogName $env(QbicCatalogName)
        }
    } else {
        set catalogName $QbicCatalogName
    }
 

    # Set the default thumbnail path
    if ![info exists QbicThumbPath] {
        if [catch {![info exists $env(QbicThumbPath)]} rc] {
            if {[string compare [string range $rc  0 9] "can't read"] == 0} {
                set thumbPath ""
            }
        } else {
            set thumbPath $env(QbicThumbPath)
        }
    } else {
	set thumbPath $QbicThumbPath
    }
    
}

# ---------PROCEDURES----------

# Move up the directory tree and update list

proc BackUp {} {
    global basePath 
    set basePath [file dirname $basePath]
    ShowListBox $basePath
}


# Display directory tree in a listbox

proc ShowDirList { currentDir } {
    global basePath
    
    # Empty the list
    .toplev.dtw.dirs delete 0 end

    # Insert an entry for the parent directory
    .toplev.dtw.dirs insert end  ../

    # Search for directories
    foreach h [lsort [glob -nocomplain -- *]] {
        if [file isdirectory $h] {
            set direntry [file tail $h]
            .toplev.dtw.dirs insert end $direntry
        }
    }
}


# Read the directory content and display in a listbox

proc ShowListBox { startDir } {

    global imageExt

    set listcnt 0

    # Guard against bogus file names
    if [catch {cd $startDir} err] {
        PathErrMsg $err
        return
    }

    # Update the base image path entry widget to display tail
    .toplev.topw.bp xview moveto 1    

    # Update the directory tree list box
    ShowDirList $startDir

    # Delete the content of the list box
    .toplev.lbw.files delete 0 end

    # Search for image files in the directory
    set j [expr [llength $imageExt]-1]
    while {$j >= 0} {
        set ext [lindex $imageExt $j]
        set ext "*.$ext"

        foreach i [lsort [glob -nocomplain -- *]] {
            if [string match $ext $i] {
                set imagefile [file tail $i]
                # Put them in the listbox
                .toplev.lbw.files insert end $imagefile
                incr listcnt 1
            }
        }
    incr j -1
    }
    if {$listcnt == 0} {
      .toplev.lbw.files insert end "No image file(s) found."
      return
    }
}

proc Cancel {} {
    # Clean up procedure goes here
    exit
}


proc Stop {} {
    dialog .d {User Canceled Script} "Script execution aborted by user!" \
    Cancel -1 OK
    exit
}

proc dialog {w title text bitmap default args} {
    global button

    # Create the top-level window and divide it into top
    # and bottom parts.

    toplevel $w -class Dialog
    wm title $w $title
    wm iconname $w Dialog
    frame $w.top -relief raised -bd 1
    pack $w.top -side top -fill both
    frame $w.bot -relief raised -bd 1
    pack $w.bot -side bottom -fill both

    # Fill the top part with the bitmap and message.

    message $w.top.msg -width 6i -text $text\
                 -font -Adobe-Times-Medium-R-Normal-*-180-*
    pack $w.top.msg -side right -expand 1 -fill both\
                 -padx 3m -pady 3m
    if {$bitmap != ""} {
        label $w.top.bitmap -bitmap $bitmap
        pack $w.top.bitmap -side left -padx 3m -pady 3m
    }

    # Create a row of buttons at the bottom of the dialog.

    set i 0
    foreach but $args {
        button $w.bot.button$i -text $but -command "set button $i"
        if {$i == $default} {
            frame $w.bot.default -relief sunken -bd 1
            raise $w.bot.button$i
            pack $w.bot.default -side left -expand 1 -padx 3m -pady 2m
            pack $w.bot.button$i -in $w.bot.default \
                    -side left -padx 2m -pady 2m -ipadx 2m -ipady 1m
        } else {
            pack $w.bot.button$i -side left -expand 1\
                  -padx 3m -pady 3m -ipadx 2m -ipady 1m
          }
          incr i
    }

    # Set up a binding for <Return>, if there`s a default,
    # set a grab, and claim the focus too.

    if {$default >= 0} {
           bind $w <Return> "$w.bot.button$default flash; \
                   set button $default"
    }
    set oldFocus [focus]
    grab set $w
    focus $w

    # Wait for the user to respond, then restore the focus
    # and return the index of the selected button.

    tkwait variable button
    destroy $w
    focus $oldFocus
    return $button
}


proc PathErrMsg { txtmsg } {
    dialog .d {Path Error} $txtmsg warning -1 OK
}


# Help screens
proc HelpAbout {} {
    toplevel .helpw
    wm title .helpw "About QbCreateCat"
    wm transient .helpw .toplev
    set topx [expr [winfo rootx .toplev] + 260]
    set topy [expr [winfo rooty .toplev] + 200]
    wm geometry .helpw +$topx+$topy
    message .helpw.ab1 -justify left -relief raised -bd 4 -aspect 1000 \
    -font -Adobe-Helvetica-Medium-R-Normal--*-240-* -text "QbCreateCat" 
    message .helpw.ab2 -justify center -relief raised -bd 1 -aspect 1000 \
    -font -Adobe-Helvetica-Medium-R-Normal--*-140-* \
    -text "Version 1.2 July 30,1996 \nAIX Tcl:v7.5a2 Tk:v4.1a2"
    message .helpw.ab3 -justify center -relief raised -bd 1 \
    -font -Adobe-Helvetica-Medium-R-Normal--*-180-* \
    -text "QbCreateCat is a Tcl/Tk-based user\
    interface for populating a searchable image database for QBIC."
    button .helpw.ok -text OK -command {destroy .helpw }
    pack .helpw.ab1 .helpw.ab2 .helpw.ab3 .helpw.ok -side top
}



proc HelpUsage {} {
    toplevel .helpu
    wm title .helpu "Using QbCreateCat"
    wm transient .helpu .toplev
    #pack .helpu -side top -fill both -expand true

    set t [text .helpu.t -setgrid true -wrap word -width 60 -height 40 \
          -yscrollcommand ".helpu.sy set"]
    scrollbar .helpu.sy -orient vert -command ".helpu.t yview"


    $t tag configure sepa -spacing1 0.2i -spacing2 0.2 -lmargin1 0.1i -rmargin 0.1i 
    $t tag configure sepahi -spacing1 0.2i -spacing2 0.2 -lmargin1 0.1i -rmargin 0.1i -font *-times-bold-r-*-20-*
    $t tag configure hilite -spacing1 0.2i -lmargin1 0.1i -lmargin2 0.1i\
                            -rmargin 0.1i -font *-times-bold-i-*-17-*
    $t tag configure para -spacing1 0.1i -lmargin1 0.4i -lmargin2 0.1i -rmargin 0.1i 
    $t tag configure hang -lmargin1 0.1i -lmargin2 0.5i 
    #$t tag configure block -spacing1 0.1i -lmargin1 0.1i -lmargin2 0.1i -rmargin 0.1i
    $t tag configure ordlst -spacing1 0.1i -lmargin1 0.1i -lmargin2 0.4i -rmargin 0.1i 

    $t insert end "Overview:\n"
    $t insert end "QbCreateCat.tcl is a Tcl script that lets you create or\
modify a QBIC catalog.  It invokes programs QbMkDbs, QbMkThmb, and indexDBM.\n"
    $t insert end "QbCreateCat.tcl reads images from the directory specified by\
'Image Directory:', and whose file extensions are given in the 'Image Extensions'\
field.  For each image, it computes the features selected by the checkboxes in\
the 'Feature Class:' list, and stores them in the database/catalog specified in\
'Database Name:' and 'Catalog Name:' fields.  To generate image thumbnails, select\
the 'Yes' radio button and specify the directory in which they are to be stored.\n"
    $t insert end "Notes:\n"
    $t insert end "1. If QbTextFeatureClass is selected, the environment variable QbicTextPath should be set.\n" 
    $t insert end "2. Computed features are inserted into the database/catalog as\
a pair consisting of (image_name, computed_features).  The image name can be fully\
qualified or not.  You control this using the 'Use Short Keys' check button.  Fully\
qualified names avoid conflicts between identically named image files in different\
directories.  Short keys (unqualified names) allow databases/catalogs and image collections\
to be moved to different directories and different systems, and allow easier portability.\n"

    $t insert end {Invocation: QbCreateCat.tcl [defaults file path]}
    $t insert end "\n"
    $t insert end {The optional defaults file '.qbictcl' may be used to set values\
for default parameters such as QbicImageBase, QbicDataPath, and QbicCatalogName.\
The current directory is the default search path for this file.  Make certain the PATH\
environment variable includes the path to the executables QbMkDbs, QbMkThmb, and indexDBM.\
The standard output and error results can be viewed in the log file /tmp/QbCreateCat.log.}
    $t insert end "\n"

    $t tag add sepahi 1.0 1.end
    $t tag add hilite 2.0 2.end
    $t tag add para 3.0 3.end
    $t tag add sepahi 4.0 4.end
    $t tag add ordlst 5.0 6.end
    $t tag add sepa 7.0 7.end
    $t tag add para 8.0 8.end

    button .helpu.ok -text OK -command {destroy .helpu}
    pack .helpu.ok -side bottom
    pack .helpu.sy -side right -fill y
    pack .helpu.t -side top -fill both -expand true

}


# View (browse) the selected file

proc Browse {imagefile} {
    global viewer basePath

    # puts $imagefile
    if  {{[ string compare $basePath "."]} !=0} {
        set imagefile $basePath/$imagefile }
    if [file isfile $imagefile] {
        exec $viewer $imagefile &
    } else {
	if [file isdirectory $imagefile] {
	    set basePath $imagefile
	    ShowListBox $basePath
	} else {
	    dialog .d {Invalid File} \
		"Unable to view $imagefile.  Check the file format." \
		error -1 OK
	    return
	}
      }
}


# Change to the selected directory and update base path

proc ChangeDir {newpath} {
    global basePath

    if {$newpath == "../"} { 
        BackUp 
    } else {
        set basePath $basePath/$newpath
        ShowListBox $basePath
    }
}


# Process the entries and selections 

proc DoIt {} {
    global logfile shkeys basePath imageExt dbName catalogName genThumb\
           thumbPath feat1 feat2 feat3 feat4 feat5 FeedBack env


   # Check for empty fields. Do not process until filled.

    if {[string compare $basePath {}] == 0} {
	 dialog .d {Incomplete Field} \
	 "Image Directory must be specified" warning -1 OK
         return
    }

    if {[string compare $imageExt {}] == 0} {
	 dialog .d {Incomplete Field} \
    "The file extension(s) of images to search for must be specified" \
         warning -1 OK
	 return
    }

    if {[string compare $dbName {}] == 0} {
	dialog .d {Incomplete Field} \
	"The database name must be specified" warning -1 OK
	return
    }

    if {[string compare $catalogName {}] == 0} {
        dialog .d {Incomplete Field} \
	"The catalog name must be specified" warning -1 OK
	return
    }

    if { [string match yes $genThumb] &&  [string compare $thumbPath {}]== 0 } {
	dialog .d {Incomplete Field} \
	"The thumbnails directory must be specified" warning -1 OK
	return
    }

    # Check for the existence of the thumbnails directory
    if { [string match yes $genThumb] && ![file exists $thumbPath] && ![file isdirectory $thumbPath]}  {
        dialog .d {Path Not Found} \
        "The thumbnails directory '$thumbPath' does not exist.
Please specify a valid path." warning -1 OK
        return
    }


    # Track number of features selected
    set featurecnt 0

    # Make sure the variable exists
    set featurelist ""

    # Build the list of features to process
    if {$feat1 == "1"} {
        set featurelist "-f QbColorFeatureClass"
        incr featurecnt
    }         
    if {$feat2 == "1"} {
        set featurelist "$featurelist -f QbTextureFeatureClass"
        incr featurecnt
    }
    if {$feat3 == "1"} {
        set featurelist "$featurelist -f QbColorHistogramFeatureClass"
        incr featurecnt
    }
    if {$feat4 == "1"} {
        set featurelist "$featurelist -f QbDrawFeatureClass"
        incr featurecnt
    }
    if {$feat5 == "1"} {
        set featurelist "$featurelist -f QbTextFeatureClass"
        incr featurecnt
    }


    # Check for "no action" - no features selected and no thumbnail generation
    if { ($featurecnt == 0) && [string match no $genThumb]} {
        dialog .d {No Action Specified} \
        "Please select a feature class to populate
database with and/or generate thumbnails." \
        warning -1 OK
        return
    }
        
    
    # No fully qualified path if 'short keys' is selected
    if {$shkeys == "1"} {
	set fileprfx ""
    } else {
	set fileprfx "$basePath/"
    }


    # Guard against bogus file names
    if [catch {cd $basePath} err] {
        PathErrMsg $err
        return
    }


    # Open the log file for writing with permission bits 0777
    # This is used by the catch commands below (QbMkDbs executions)
    set logfileID [open $logfile w 0777]


    # If a non-qualified thumbpath is entered, prepend base image path
    if [string match $thumbPath [file tail $thumbPath]] {
        set thumbPath $basePath/$thumbPath
    }
    # Set the shell environment variables QbicThumbPath &
    # QbicImagePath required by QbMkThmb
    set env(QbicThumbPath) $thumbPath
    set env(QbicImagePath) $basePath


    # Eval forces another level of parsing and so $featurelist
    # is translated into multiple arguments, not a single argument string.


    # If no features were selected make no call(s) to QbMkDbs
    if {$featurecnt > 0} {

        set Qbargs "-d $dbName -c $catalogName $featurelist"
        if [catch {[eval exec QbMkDbs $Qbargs >& $logfile]} rc1] {
            if {[string compare $rc1 "child process exited abnormally"] == 0} {
                puts $logfileID "Initial catalog creation failed.  RC=$rc1."
            } else {
                puts $logfileID "$catalogName created successfully."
            }
        } 
    }

    # Files processed counter
    set filecnt 0

    # Counter for number of files failed to insert into the database
    set failedcnt 0


    set nExten [expr [llength $imageExt]-1]
    while {$nExten >= 0} {
        set ext [lindex $imageExt $nExten]

        foreach n [lsort [glob nocomplain -- *]] {

            if [regexp $ext $n] {
		# DEBUG: What if the user enters ?+* ?

                # Truncate the path, retain just the file name
                set imagefile [file tail $n]


                # If any feature is selected, execute QbMkDbs
                if {$featurecnt > 0} {
                    # Append resulting standard error to log file
                    puts stdout "exec QbMkDbs -d $dbName -c $catalogName $fileprfx$imagefile"
                    if [catch {[exec QbMkDbs -d $dbName -c $catalogName $fileprfx$imagefile >>& $logfile]} rc2] {
                        if {[string compare $rc2 "child process exited abnormally"] == 0} {
                            puts $logfileID "Insert failed for file: $fileprfx$imagefile"
                            incr failedcnt 1
                        }
                    } else {
                        # Inserted successfully above and stdout is written to logfile
                    }
                }


                incr filecnt 1

                # Create the thumbnail of the image file
                if [string match yes $genThumb] {
                    puts stdout "exec QbMkThmb $imagefile $thumbPath/$imagefile.thumb0.gif"
		    set rc3 [eval exec QbMkThmb $imagefile $thumbPath/$imagefile.thumb0.gif >>& $logfile]
                }
            }
        }
    incr nExten -1
    }

    # Post processing - create the index file from a feature database

    # The path to "QbicData" has been assigned above to dbName and it's fully qualified
    if [catch {cd $dbName} err] { 
        PathErrMsg $err
        puts $logfileID $err
        return
    }
    
    set dbmf {}
    foreach m [glob nocomplain -- Color*.$catalogName Draw*.$catalogName Text*.$catalogName ] {
        lappend dbmf $m
    }

    set idxfile [lindex $dbmf 0]

    if { $idxfile != {}} {
        set rc [eval exec indexDBM $idxfile >>& $logfile ] 
        puts stdout "exec indexDBM $idxfile"
        set rc [eval exec mv $idxfile.index.dbm $catalogName.index]  
        puts stdout "exec mv $idxfile.index.dbm $catalogName.index"
    } else {
        puts $logfileID "No index file created."
    }

    close $logfileID

    set succ [expr $filecnt-$failedcnt]

    if {$featurecnt == 0} {
        set succstr ""
    } else {
        set succstr "  $succ inserted successfully."
    }

    dialog .d {Script Complete} \
      "Processed $filecnt file(s).$succstr" info -1 OK

    exit
}

# --------WIDGETS CREATION--------
# Don't show the shell level window
wm withdraw .

# Define a top level widget
toplevel .toplev

# Place the window in the center
# wm geometry .toplev +270+160

# Set window title
wm title .toplev QbCreateCat

# Top file menu bar
frame .toplev.mbframe -bg grey -relief raised
pack .toplev.mbframe -side top -fill x


# Help menu bar
menubutton .toplev.mbframe.hmenu -text Help -underline 0 -menu .toplev.mbframe.hmenu.help
set hm [menu .toplev.mbframe.hmenu.help -tearoff 0]
$hm add command -label "About QbCreateCat..." -command HelpAbout
$hm add separator
$hm add command -label "Usage" -command HelpUsage

$hm config -fg black -bg grey
.toplev.mbframe.hmenu config -fg black -bg grey

pack .toplev.mbframe.hmenu -side right

# Frame for buttons and entry
frame .toplev.topw -borderwidth 4
pack .toplev.topw -side top -fill x
.toplev.topw config -cursor {hand1 white black}

# backUp climbs one level up the directory tree
button .toplev.topw.back -text Up -command BackUp
# Button to update the list box
button .toplev.topw.search -text "List Images" -command {ShowListBox $basePath}
pack .toplev.topw.search .toplev.topw.back -side right

# Labeled entry for the base search directory path
label .toplev.topw.bl -text "Image Directory:" -padx 0
entry .toplev.topw.bp -width 30 -relief sunken -textvariable basePath
pack .toplev.topw.bl -side left
pack .toplev.topw.bp -side left -fill x -expand true
.toplev.topw.bp xview moveto 1
.toplev.topw.bp config -fg white -bg #4A766E

# *** Global variable: basePath

# Set up key binding equivalents to the buttons
bind .toplev.topw.bp <Return> {ShowListBox $basePath}
bind .toplev.topw.bp <Control-b> BackUp
bind .toplev.topw.bp <Control-c> Stop
focus .toplev.topw.bp        

# Frame for image file extensions entry widget
frame .toplev.ifew -borderwidth 2
pack .toplev.ifew -side top -fill x

# Labeled entry for file extensions specification
label .toplev.ifew.fl -text "Image File Extensions:" -padx 0
entry .toplev.ifew.fext -width 20 -relief sunken -textvariable imageExt
pack .toplev.ifew.fl -side left -fill x
pack .toplev.ifew.fext -side left -fill x -expand true
.toplev.ifew.fext config -fg white -bg #4A766E

# *** Global variable: imageExt


# List box for listing directory tree
frame .toplev.dtw
label .toplev.dtw.dtl -text "Directories" -padx 0 -pady 0
listbox .toplev.dtw.dirs -width 15 -relief groove -borderwidth 2 \
          -yscrollcommand ".toplev.dtw.scrolly set" \
          -xscrollcommand ".toplev.dtw.scrollx set"
scrollbar .toplev.dtw.scrolly -orient vertical \
          -command ".toplev.dtw.dirs yview"
scrollbar .toplev.dtw.scrollx -orient horizontal \
          -command ".toplev.dtw.dirs xview"
pack .toplev.dtw .toplev.dtw.dirs .toplev.dtw.dtl -side top

pack append .toplev.dtw .toplev.dtw.scrolly {left filly} \
    .toplev.dtw.scrollx {bottom fillx} \
    .toplev.dtw.dirs {expand fill}
pack .toplev.dtw .toplev.dtw.dirs -side left -fill both -expand true 

bind .toplev.dtw.dirs <Double-Button-1> {
    foreach d [selection get] { ChangeDir $d } }
focus .toplev.dtw.dirs
.toplev.dtw config -cursor {crosshair black white}
.toplev.dtw.dirs config -fg white -bg #4A766E


# List box widget listing files in directory
frame .toplev.lbw
label .toplev.lbw.sdl -text "Images Found" -padx 0 -pady 0
listbox .toplev.lbw.files -width 15 -relief groove -borderwidth 2 \
         -yscrollcommand ".toplev.lbw.scrolly set" \
        -xscrollcommand ".toplev.lbw.scrollx set"
#pack .toplev.lbw.files -side top -expand 1 -fill x
scrollbar .toplev.lbw.scrolly -orient vertical \
             -command ".toplev.lbw.files yview"
scrollbar .toplev.lbw.scrollx -orient horizontal \
             -command ".toplev.lbw.files xview"
pack .toplev.lbw .toplev.lbw.files .toplev.lbw.sdl -side top 
pack append .toplev.lbw .toplev.lbw.scrolly {right filly} \
    .toplev.lbw.scrollx {bottom fillx} \
    .toplev.lbw.files {expand fill}
pack .toplev.lbw .toplev.lbw.files -side left -fill both -expand true 
bind .toplev.lbw.files <Double-Button-1> {
    foreach f [selection get] { Browse $f } }
focus .toplev.lbw.files
.toplev.lbw config -cursor {crosshair black white}
.toplev.lbw.files config -fg white -bg #4A766E


# Short keys button
set shkeys 0
frame .toplev.skw -borderwidth 2
pack .toplev.skw -side top -fill x
.toplev.skw config -cursor {hand1 white black}
label .toplev.skw.skl -text "Filenames Qualifier" 
checkbutton .toplev.skw.skbut -text "Use Short Keys" -variable shkeys -anchor w
pack .toplev.skw.skl -side top
pack .toplev.skw.skbut -side top
.toplev.skw.skbut config -fg white -bg #4A766E


# Frame for database name entry widget
frame .toplev.dnw -borderwidth 4
pack .toplev.dnw -side top -fill x 

# Labeled entry for database name
label .toplev.dnw.dl -text "Database Name:"
entry .toplev.dnw.dbn -width 20 -relief sunken -textvariable dbName
pack .toplev.dnw.dl -side top
pack .toplev.dnw.dbn -side right -fill x -expand true
.toplev.dnw.dbn xview moveto 1
.toplev.dnw.dbn config -fg white -bg #4A766E


# *** Global variable: dbName

# Frame for catalog name entry widget
frame .toplev.cnw -borderwidth 2
pack .toplev.cnw -side top -fill x
# Labeled entry for catalog name
label .toplev.cnw.cl -text "Catalog Name:"
entry .toplev.cnw.catn -width 20 -relief sunken -textvariable catalogName
pack .toplev.cnw.cl -side top
pack .toplev.cnw.catn -side top
.toplev.cnw.catn config -fg white -bg #4A766E

# *** Global variable: catalogName

# Set check buttons to activate all features initially
set feat1 1
set feat2 1
set feat3 1
set feat4 1
set feat5 1
# Frame for feature class buttons widget
frame .toplev.fcw -borderwidth 4
pack .toplev.fcw -side top
.toplev.fcw config -cursor {hand1 white black}
# Feature class check buttons
label .toplev.fcw.fcl -text "Feature Class:"
checkbutton .toplev.fcw.colfeatbut -text QbColorFeatureClass -variable feat1 -anchor w
.toplev.fcw.colfeatbut config -fg white -bg #4A766E
checkbutton .toplev.fcw.txtrfeatbut -text QbTextureFeatureClass -variable feat2 -anchor w
.toplev.fcw.txtrfeatbut config -fg white -bg #4A766E
checkbutton .toplev.fcw.histfeatbut -text QbColorHistogramFeatureClass -variable feat3 -anchor w
.toplev.fcw.histfeatbut config -fg white -bg #4A766E
checkbutton .toplev.fcw.drawfeatbut -text QbDrawFeatureClass -variable feat4 -anchor w
.toplev.fcw.drawfeatbut config -fg white -bg #4A766E
checkbutton .toplev.fcw.textfeatbut -text QbTextFeatureClass -variable feat5 -anchor w
.toplev.fcw.textfeatbut config -fg white -bg #4A766E
pack .toplev.fcw.fcl .toplev.fcw.colfeatbut .toplev.fcw.txtrfeatbut .toplev.fcw.histfeatbut \
     .toplev.fcw.drawfeatbut .toplev.fcw.textfeatbut -side top -fill x

# *** Global variables (binary): feat1, feat2, feat3, feat4, feat5

# Set default button to yes
set genThumb yes
# Frame for Generate Thumbnail radio buttons
frame .toplev.gtw -borderwidth 10
pack .toplev.gtw -side top
.toplev.gtw config -cursor {hand1 white black}
# Generate Thumbnail radio buttons
label .toplev.gtw.gtl -text "Generate Thumbnails:"
radiobutton .toplev.gtw.gtyes -text Yes -variable genThumb -value yes -anchor w
.toplev.gtw.gtyes config -fg white -bg #4A766E
radiobutton .toplev.gtw.gtno -text No -variable genThumb -value no -anchor w
.toplev.gtw.gtno config -fg white -bg #4A766E
pack .toplev.gtw.gtl .toplev.gtw.gtyes .toplev.gtw.gtno -side top -fill x


# Flag to enable/disable the entry widget
# genThumb is defaulted to yes and the above widget is radiobuttons.
# Future modification: convert .gtw.gtyes & .gtw.gtyes radiobuttons into buttons.
if { [string compare $genThumb "yes"] == 0 } {
   set entryState normal
} else {
    set entryState disabled
}

                     
# *** Global variable: genThumb

# Entry widget for thumbnail path
label .toplev.gtw.gdl -text "Directory for Thumbnails:"
entry .toplev.gtw.gtdir -state $entryState -width 20 -relief sunken -textvariable thumbPath
pack .toplev.gtw.gdl .toplev.gtw.gtdir -side top -fill x
.toplev.dnw.dbn xview moveto 1
.toplev.gtw.gtdir config -fg white -bg #4A766E


# *** Global variable: thumbPath

# Frame for the command buttons
frame .toplev.cbw -borderwidth 2
pack .toplev.cbw -side top -padx 1 -pady 1
button .toplev.cbw.ok -relief raised -text "  OK  " -command  DoIt
button .toplev.cbw.cancel -relief raised -text Cancel -command Cancel
pack .toplev.cbw.ok .toplev.cbw.cancel -side left \
    -padx 1m -pady 1m -ipadx 1m -ipady 1m -expand true
.toplev.cbw config -cursor {hand1 white black}


#EOF
